/*
 * ConnectedSubgraph.java
 *
 * Copyright (c) 2006-2007 Computer-Aided Integrated Systems Laboratory, 
 * St. Petersburg Institute for Informatics and Automation RAS http://cais.lisa.iias.spb.su
 *
 * This software is open-source under the BSD license; see either "license.txt" 
 * or http://jung.sourceforge.net/license.txt for a description.
 */

package connectedsubgraph;

import matching.GraphComparator;
import matching.VertexPairSimilar;
import matching.UnweightedShortestPathUtil;
import edu.uci.ics.jung.algorithms.shortestpath.UnweightedShortestPath;

import edu.uci.ics.jung.utils.UserData;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.Edge;
import edu.uci.ics.jung.graph.impl.SparseVertex;
import edu.uci.ics.jung.graph.impl.SparseGraph;
import edu.uci.ics.jung.graph.impl.UndirectedSparseEdge;

import java.util.Map;
//import java.util.HashMap;
import java.util.List;
//import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.Comparator;


/** Selects connected subgraph
 *
 */
public class ConnectedSubgraph {
    
    /** Update data about neighbours sets of vertices (neighbours to the set 'c'),
     *  stored in the DistanceComponent cc_neighbours_all.
     */
    static public void updateNeighbours(Set<Vertex> c,
            DistanceComponent dc_merging,
            Set<Vertex> c_new,
            Map<Set<Vertex>, DistanceComponent[]> cc_neighbours_all, 
            List<DistanceComponent> dc_all,
            UnweightedShortestPath usp_all) {
            
        for(DistanceComponent d_neighbour:cc_neighbours_all.get(c)) {
            if(d_neighbour != dc_merging) {
                assert(d_neighbour.cc1 == c || 
                       d_neighbour.cc2 == c);
                Set<Vertex> c_neighbour = d_neighbour.cc1 == c ? d_neighbour.cc2 : d_neighbour.cc1;

                // update fields of DistanceComponent d_neighbour for sets 'c' and 'c_neighbour'
                Vertex[] path = UnweightedShortestPathUtil.getShortestPath(c, c_neighbour, usp_all);
                if(null != path && 0 < path.length) {
                    d_neighbour.update(c, c_neighbour, path);
                } else {
                    assert(1==0); // it's impossible that there is no distance between two sets of vertices!
                    c_new.addAll(c_neighbour);  // thus, this is unreachable line, I hope.
                    dc_all.remove(d_neighbour); // Else we 'll need call getShortestPath for each in 'dc_all' recursively.
                }
            }
        }
    }
    
    public static final Comparator<DistanceComponent> PATH_DIST_ORDER = new Comparator<DistanceComponent>() {
        public int compare(DistanceComponent d1, DistanceComponent d2) {
            if (d1.path.length > d2.path.length)
                return -1;
            return 1;
        }
    };
    /** Selects minimum subset of connected vertices from the graph 'g_all' 
     * by set of initial vertices in the graph 'g_cur'. Returns the subgraph.
     *
     * It is possible that vertices from the initial graph 'g_cur' 
     * are not connected directly in the graph 'g_all'. 
     * If they are connected via some other vertices, then these vertices 
     * will be added to the result graph 'g_cur'.
     * <PRE>
     * 1. Select graph connected components (CC), store to the array 'cc'
     *    Depth-first search (DFS)
     * 2. Find shortest paths between CC, 
     *    (shortest path contains vertices from the set 'g_all/g_cur')
     *      store to the array 'path', 
     *      sort arrays by 'path.length' (path[0].length is minimum)
     * 3. Add vertices and edges from 'path[0]' to graph 'g_cur', (TODO! TO dif. function)
     *    (e.g. it is a shortest path between components CC1 and CC2).
     *    Merge CC1 and CC2 into CC1, delete CC2 from array of componets.
     *    Update 'path' and 'dist' including CC1 or CC2.</PRE>
     *
     * @param vert_subgraph_to_vert map of subgraph vertices (g_cur) to vertices of the whole graph (g_all)
     * @param vert_to_vert_subgraph map of vertices of the whole graph (g_all) to vertices of the subgraph (g_cur)
     *
     * Attention!: vertices labels in the graph should be unique.
     */
    public static Graph selectsMinimumVerticesSetByLabels (
            Graph g_cur,Graph g_all,
            Map<Vertex, Vertex> vert_subgraph_to_vert,
            Map<Vertex, Vertex> vert_to_vert_subgraph) {
     
        // 1
        ConnectedComponent concom = new ConnectedComponent();
    
        // array of connected components of g_cur in g_all
        List<Set<Vertex>> cc_cur = concom.getConnectedComponents(g_cur);
        List<Set<Vertex>> cc_all = ConnectedComponent.mapSetToSet(cc_cur, vert_subgraph_to_vert);
        if(null == cc_cur || 1 == cc_cur.size()) {
            return g_cur;
        }
        assert(1 < cc_cur.size());
        
        // 2.
        UnweightedShortestPath usp_all = new UnweightedShortestPath(g_all);
        List<DistanceComponent> dc_all = DistanceComponent.calcShortestDistances(cc_all, usp_all);
        
        
        // sort dc by dist path
        // todo 
        //Collections.sort(edges, GraphComparator.SIM_ORDER);
        
        Map<Set<Vertex>, DistanceComponent[]> cc_neighbours_all = DistanceComponent.mapSetToNeighbourComponents(dc_all,cc_all);
        
        // 3.
        while (1 < cc_all.size() && 0 < dc_all.size()) {
            DistanceComponent dc_merging = dc_all.remove(0);
            
            Set<Vertex> c_new = new HashSet<Vertex>();
            c_new.addAll(dc_merging.cc1);
            c_new.addAll(dc_merging.cc2);
            
            // add vertices and edges of the shortest path
            boolean prev_sub_v2_null = false;
            for(int i=0; i<dc_merging.path.length-1; i++) {
                Vertex v1 = dc_merging.path[i];  // v1 = VertexPairSimilar.getVertexByName(g_cur, (String)dc_merging.path[i].getUserDatum( GraphComparator.NAME_KEY ));
                Vertex v2 = dc_merging.path[i+1];
                
                Vertex sub_v1 = vert_to_vert_subgraph.get(v1);
                Vertex sub_v2 = vert_to_vert_subgraph.get(v2);
                
                if(null == sub_v2) {
                    Vertex v = g_cur.addVertex(new SparseVertex());
                    String name = (String)((Vertex)v2).getUserDatum(GraphComparator.NAME_KEY);
                    v.addUserDatum(GraphComparator.NAME_KEY, name, UserData.SHARED);
                    vert_to_vert_subgraph.put(v2, v);
                    vert_subgraph_to_vert.put(v, v2);
                    
                    assert(!c_new.contains(v2));
                    c_new.add(v2);
                    
                    assert(null != sub_v1);
                    g_cur.addEdge(new UndirectedSparseEdge(sub_v1, v));
                } else {
                    if( 2   == dc_merging.path.length && 
                       null == sub_v1.findEdge(sub_v2))
                    {
                        g_cur.addEdge(new UndirectedSparseEdge(sub_v1, sub_v2));
                    }
                }
                
                if( i == dc_merging.path.length-2 && 
                    prev_sub_v2_null) 
                {
                    assert(null != sub_v2);
                    g_cur.addEdge(new UndirectedSparseEdge(sub_v1, sub_v2));
                }
                prev_sub_v2_null = null == sub_v2;
            }
            
            // update neighbours of merged set of vertices cc1, cc2
            updateNeighbours(dc_merging.cc1, dc_merging, c_new, cc_neighbours_all, dc_all, usp_all);
            updateNeighbours(dc_merging.cc2, dc_merging, c_new, cc_neighbours_all, dc_all, usp_all);
            
            cc_all.remove(dc_merging.cc1);
            cc_all.remove(dc_merging.cc2);
            cc_all.add(c_new);
            
            // sort dc by dist path
            // todo 
            // ...
        }
        
        return g_cur;
    }

    
    
    /** Creates a new instance of ConnectedSubgraph */
    public ConnectedSubgraph() {
    }
    
    
}
