/*
 * Copyright (c) 2006-2007 Computer-Aided Integrated Systems Laboratory, 
 * St. Petersburg Institute for Informatics and Automation RAS http://cais.lisa.iias.spb.su
 *
 * This software is open-source under the BSD license; see either "license.txt" 
 * or http://jung.sourceforge.net/license.txt for a description.
 */

package matching;

//import topindex.LevenshteinDistance;

import edu.uci.ics.jung.io.PajekNetReader;
import edu.uci.ics.jung.algorithms.shortestpath.UnweightedShortestPath;
import edu.uci.ics.jung.graph.decorators.Indexer;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.utils.UserData;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Collections;

        
/** Calculates the similarity between two graphs.
 *
 * The similarity between two vertices is stored in vertices of graph1, 
 * use </CODE>v.getUserDatum(SimFlooding.SIM_KEY)</CODE> to get similarity value.
 *
 */
public class GraphComparator {

    /** vertex name (field of every vertex) */
    static public final String NAME_KEY = "name";
    
    /** the vertex can belong to subgraph 1, 2, ... */
    static public final String SUBGRAPH_N = "subgraph_n";
    
    /** field of some edges, it's a similarity between vertices*/
    static public final String SIM_KEY = "sim_value";
    
    /** Vertices of graph1 similar (by name) to some vertices of graph2, Map <vertex_from_graph1, vertex_from_graph2> */
    public Map<Vertex,Vertex> similar_v1_v2;
    
    /** similar_v2_v1 is a vice versa of similar_v1_v2, Map <vertex_from_graph2, vertex_from_graph1> */
    public Map<Vertex,Vertex> similar_v2_v1;
    /** Vertices of graph2 similar (by name) to some vertices of graph1 */
    //private List<Vertex> similar_vert2;
    
    /** similarity parameters between two vertices: from similar_v1_v2 and similar_v2_v1 (for debug purpose) */
    public Map<Vertex,VertexPairSimilar> pair_similar;
    
    /** Loads labels of vertex of graph (graph g is taken from Pajek file), 
     * i.e. copies vertex user datum from PajekNetReader.LABEL to GraphComparator.NAME_KEY.
     */
    public static void reloadPajekVertexLabelsToGraph(Graph g) {
        for(Object o:g.getVertices()) {
            Vertex v = (Vertex)o;
            String s = (String)(v.getUserDatum(PajekNetReader.LABEL));
            v.setUserDatum(GraphComparator.NAME_KEY, s, UserData.SHARED);
        }
    }
    
    
    public void calcNodesStringSimilarityOld(Graph g1, Graph g2, LevenshteinDistance lev) {
        similar_v1_v2 = new HashMap<Vertex,Vertex>();
        similar_v2_v1 = new HashMap<Vertex,Vertex>();
                
        for(Object o1:g1.getVertices()) {
            Vertex v1 = (Vertex)o1;
            
            int  dist_min = Integer.MAX_VALUE;
            Vertex  v_min = null;
            for(Object o2:g2.getVertices()) {
                Vertex v2 = (Vertex)o2;
                if(!similar_v2_v1.containsKey(v2)) {
                    String name1 = (String)v1.getUserDatum(GraphComparator.NAME_KEY);
                    String name2 = (String)v2.getUserDatum(GraphComparator.NAME_KEY);
                    int dist = lev.computeLevenshteinDistance(
                                    name1,
                                    name2);
                    if(-1 != dist && dist < dist_min) {
                        dist_min = dist;
                        v_min = v2;
                    }
                }
            }
            
            if(null != v_min) {
                v1.setUserDatum(GraphComparator.SIM_KEY, (int)dist_min, UserData.SHARED);
                similar_v1_v2.put(v1,v_min);
                similar_v2_v1.put(v_min,v1);
            }
        }
    }
    
    
    
    class VertexPair {
        Vertex v1, v2;
        
        /** Distance between names of vertices v1 and v2 */
        int sim;
        
        public VertexPair(Vertex v, Vertex vv, int distance) {
            v1 = v; v2 = vv; sim = distance;
        }
    }
    class ListPairs {
        List<VertexPair> pairs;
    }
    public static final Comparator<VertexPair> SIM_ORDER = new Comparator<VertexPair>() {
        public int compare(VertexPair p1, VertexPair p2) {
            if (p1.sim <= p2.sim)
                return -1;
            return 1;
        }
    };
    
    
    /** Calculates initial similarity of node names (NAME_KEY) via Levenshtein 
     * distance between nodes of sub_graph1 and sub_graph2, stores similarity 
     * value to vertex field (SIM_KEY).
     *
     * Creates map between similar vertices in graph 1 and 2. The most similar
     * vertex in graph2 (only one vertex, i.e. one-to-one mapping) will be 
     * mapped for each vertex in graph1. The greedy algorithm.
     */
    public void calcNodesStringSimilarity(Graph g1, Graph g2, LevenshteinDistance lev) {
        
        // get list of edges sorted by length
        //Map<Vertex,ListPairs> v1_pairs = new HashMap<Vertex,ListPairs>();
        //Map<Vertex,ListPairs> v2_pairs = new HashMap<Vertex,ListPairs>();
        List<VertexPair> edges = new ArrayList<VertexPair>();
        
        // fill v1_pairs, v2_pairs, edges.
        for(Object o1:g1.getVertices()) {
            Vertex v1 = (Vertex)o1;
            for(Object o2:g2.getVertices()) {
                Vertex v2 = (Vertex)o2;
                
                String name1 = (String)v1.getUserDatum(GraphComparator.NAME_KEY);
                String name2 = (String)v2.getUserDatum(GraphComparator.NAME_KEY);
                int dist = lev.computeLevenshteinDistance(
                                name1,
                                name2);
                if(-1 != dist) {
                    VertexPair pair = new VertexPair(v1, v2, dist);
                    edges.add(pair);
                    
                    //List<VertexPair> list1 = v1_pairs.containsKey(v1) ? v1_pairs.get(v1) : new ArrayList<VertexPair>();
                    //v1_pairs.put(v1, list1);
                    //List<VertexPair> list2 = v2_pairs.containsKey(v2) ? v2_pairs.get(v2) : new ArrayList<VertexPair>();
                    //v2_pairs.put(v2, list2);
                }
            }
        }
        
        Collections.sort(edges, GraphComparator.SIM_ORDER);
        
        similar_v1_v2 = new HashMap<Vertex,Vertex>();
        similar_v2_v1 = new HashMap<Vertex,Vertex>();
        for(VertexPair e:edges) {
            if(!similar_v1_v2.containsKey(e.v1) && 
               !similar_v2_v1.containsKey(e.v2)
              ) {
                e.v1.setUserDatum(GraphComparator.SIM_KEY, e.sim, UserData.SHARED);
                similar_v1_v2.put(e.v1,e.v2);
                similar_v2_v1.put(e.v2,e.v1);
            }
        }
    }
    
    
    /** Compares two graph, returns value of similarity. Takes into account:
     *  1) Similarity between names of nodes;
     *  2) Number of neighbours of similar nodes in graph 1 & 2;
     *  3) Distance of shortest path between similar nodes in graph 1 & 2.
     *
     *  Map this.pair_similar is filled here.
     */
    public double compare(Graph g1, Graph g2, LevenshteinDistance lev) {
        
        if(null == g1 || 0 == g1.numVertices() || 
           null == g2 || 0 == g2.numVertices()
          ) {
            return 0.0;
        }
        
        // calculates initial similarity, fills similar_v1_v2, similar_v2_v1
        calcNodesStringSimilarity(g1, g2, lev);
        
//        Indexer id1 = Indexer.getIndexer(g1);
//        Indexer id2 = Indexer.getIndexer(g2);
        
        UnweightedShortestPath usp1 = new UnweightedShortestPath(g1);
        UnweightedShortestPath usp2 = new UnweightedShortestPath(g2);
        
        pair_similar = new HashMap<Vertex,VertexPairSimilar>();
        
        for(Vertex v1:similar_v1_v2.keySet()) {
            Vertex v2 = similar_v1_v2.get(v1);
            assert(null != v2);
            
            Map<Vertex,Integer> v1_path_len = UnweightedShortestPathUtil.getPath(v1, similar_v1_v2.keySet(), usp1);
            Map<Vertex,Integer> v2_path_len = UnweightedShortestPathUtil.getPath(v2, similar_v2_v1.keySet(), usp2);
            
            Map<Vertex,Integer> v1_path_len_real = UnweightedShortestPathUtil.getPathReal(v1, similar_v1_v2.keySet(), usp1);
            Map<Vertex,Integer> v2_path_len_real = UnweightedShortestPathUtil.getPathReal(v2, similar_v2_v1.keySet(), usp2);
            
            VertexPairSimilar pair = new VertexPairSimilar(v1, v2, 
                    //usp1, usp2, // delete, add v1_path_len_real
                    (Integer)v1.getUserDatum(GraphComparator.SIM_KEY), 
                    similar_v1_v2,      similar_v2_v1,
                    v1.numNeighbors(),  v2.numNeighbors(), 
                    v1_path_len,        v2_path_len,
                    v1_path_len_real,   v2_path_len_real);
            pair_similar.put(v1, pair);
        }
        
        return VertexPairSimilar.calcSimilarity(pair_similar, g1.numVertices(), g2.numVertices());
    }
    
}
