/*
 * LevenshteinDistance.java
 *
 * Copyright (c) 2006 Computer-Aided Integrated Systems Laboratory, 
 * St. Petersburg Institute for Informatics and Automation RAS http://cais.lisa.iias.spb.su
 *
 * This software is open-source under the BSD license; see either "license.txt" 
 * or http://jung.sourceforge.net/license.txt for a description.
 *
 * See:  http://en.wikisource.org/wiki/Levenshtein_distance#Java
 *       http://en.wikipedia.org/wiki/Levenshtein_distance
 */

package matching;

import java.util.*;

/** Compare words with skipping stop words.
 * The distance is less for more similar words, it is 0 for identical words.
 * Returns -1, if use_dist_matrix is true and if words are not similar.
 */
public class LevenshteinDistance {
    
    static final String[] array_stop_words = {"and", "or", "units"};
    public HashSet<String> stop_words;
    
    public boolean use_dist_matrix;
    
    /** Distance matrix defines maximum distance between words, when to start consider them as unsimilar, 
     * e.g. <code>constraint[i] = j</code> means that <code>word.length==i</code> and maximum allowed 
     * distance is j-1. */
    int []  constraint;
    int     max_dist;
    
    static final int min_length = 3;    // The words which are equal or shorter than min_length will not be compared.
    
    /** Soft constraints defines more strict conditions in order to consider two words to be similar */
    public static final short TYPE_SOFT = 1;
    /** Hard constraints defines less strict conditions in order to consider two words to be similar */
    public static final short TYPE_HARD = 0;
    private int  constraint_type;
    
    public int getMaxDist() {
        return max_dist;
    }
            
    /** Defines level of strict conditions in order to consider two words to be similar.
     * If <code>use_dist_matrix</code> is false then this function has no effect, since
     * distance matrix (constraint[]) will not be used, set use_dist_matrix=true before.
     */
    public void setContstraintType(short i) {
        switch (i) {
            case TYPE_SOFT:
                constraint = soft_constraint;
                max_dist = soft_max_dist;
                break;
            case TYPE_HARD:
                constraint = hard_constraint;
                max_dist = hard_max_dist;
                break;
            default:
                System.out.println("Warning. Unknown constraint type:"+i);
        }
    }
    
            
    // Experimental values
    
    // soft table
    static final int [] soft_constraint = {
     // maximum distance 
            //  words length        
            //  words length = len(test word)
        0,  //  0                   // :test word      :similar words(Distance)
        0,  //  1
        0,  //  2
        2,  //  3
        3,  //  4
        4,  //  5
        5,  //  6                   //  :Safety         :Supply(4), Market(4), Part(4)
        6,  //  7
        7,  //  8                   //  :Disaster       :Dealer(4)
        7,  //  9
        8,  //  10
        8,  //  11
        9,  //  12
        9   //  13
    };
    static final int soft_max_dist   = 10;
    
    // "hard" strict table
    static final int [] hard_constraint = {
     // maximum distance 
            //  words length        
            //  words length = len(test word)
        0,  //  0                   // :test word      :similar words(Distance)
        0,  //  1
        0,  //  2
        2,  //  3
        2,  //  4
        2,  //  5
        3,  //  6                   //  :Safety         :Supply(4), Market(4), Part(4)
        3,  //  7
        3,  //  8                   //  :Disaster       :Dealer(4)
        4,  //  9
        4,  //  10
        4,  //  11
        5,  //  12
        5   //  13
    };
    static final int hard_max_dist   = 6;   // max_dist is a distance for words with length >= constraint.length, then words are not similar
     
    
    public LevenshteinDistance() {
        // initialize stop_words
        stop_words = new HashSet<String>();
        for(String s:array_stop_words) {
            stop_words.add(s);
        }
        setContstraintType(LevenshteinDistance.TYPE_HARD);
    }
    
    private static int minimum(int a, int b, int c)
    {
        if (a<=b && a<=c)
            return a;
        if (b<=a && b<=c)
            return b;
        return c;
    }
    
    
    private static int computeLevenshteinDistance(char[] str1, char[] str2) {
        int[][] distance = new int[str1.length+1][];

        for(int i=0; i<=str1.length; i++)
        {
            distance[i] = new int[str2.length+1];
            distance[i][0] = i;
        }
        for(int j=0; j<str2.length+1; j++)
            distance[0][j]=j;

        for(int i=1; i<=str1.length; i++)
            for(int j=1;j<=str2.length; j++)
                distance[i][j]= minimum(distance[i-1][j]+1,
                    distance[i][j-1]+1, distance[i-1][j-1] + 
                    ((str1[i-1]==str2[j-1])?0:1));
       
        return distance[str1.length][str2.length];
    }
    
    public int computeLevenshteinDistance(String w1, String w2) {
        int  l1, l2, len;
        if(null == w1 || null == w2) {
            return -1;
        }
        int d = computeLevenshteinDistance(w1.toCharArray(), w2.toCharArray());
        
        if(use_dist_matrix) {
            // Check by constraint table whether the word is similar to another one,
            // using the word length and distance between words
            l1 = w1.length();
            l2 = w2.length();
            len = l1 < l2 ? l1 : l2;
            if((len >= constraint.length && d >= max_dist) ||
               (len <  constraint.length && d >= constraint[len]) ||
                stop_words.contains(w1) ||
                stop_words.contains(w2))
            {
                d = -1;
            }
        }
        return d;
    }
    
    
    /** Compare words with skipping stop words.
     * Return -1, if use_dist_matrix is true and if words are not similar, 
     * (i.e. word's length is equal or shorter than min_length or 
     * distance between words is bigger than max_dist).
     * Return -1, if one of words is null or "".
     */
    public int computeLevenshteinDistance(String[] words1, String[] words2) {
        int d, distance, l1, l2, len;
        
        String best1="", best2="";
        boolean b=false;
        
        distance = Integer.MAX_VALUE;
        for(String w1:words1) {
            if(null != w1 && min_length < w1.length() && !stop_words.contains(w1) ) {
                for(String w2:words2) {
                    if(null != w2 && min_length < w2.length() && !stop_words.contains(w2)) {
                        d = computeLevenshteinDistance(w1.toCharArray(), w2.toCharArray());
                        if(d < distance) {
                            
                            // Check by constraint table whether the word is similar to another one,
                            // using the word length and distance between words
                            l1 = w1.length();
                            l2 = w2.length();
                            len = l1 < l2 ? l1 : l2;
                            if(!use_dist_matrix ||
                                (len >= constraint.length && d < max_dist) ||
                                (len <  constraint.length && d < constraint[len])) 
                            {
                                    distance = d;
                                    best1 = w1;
                                    best2 = w2;
                                    b = true;
                            }
                        }
                    }
                }
            }
        }
        /*if(b) {
            System.out.println(distance+"=Levenshtein('"+best1+"','"+best2+"')");
        }*/
        
        if(use_dist_matrix && distance > max_dist) {
            distance = -1;  
        }
        
        return distance;
    }

    
    /*
    public boolean wordIsSimilar(int len, int distance) { 

        if(!use_dist_matrix ||
                (len >= max_length && distance < max_dist) ||
                (len <  max_length && distance < constraint[len]))
        {
                return true;
        }
        return false;
    }*/
}
