/*
 * Copyright (c) 2006-2007 Computer-Aided Integrated Systems Laboratory, 
 * St. Petersburg Institute for Informatics and Automation RAS http://cais.lisa.iias.spb.su
 *
 * This software is open-source under the BSD license; see either "license.txt" 
 * or http://jung.sourceforge.net/license.txt for a description.
 */

package matching;

import edu.uci.ics.jung.algorithms.shortestpath.UnweightedShortestPath;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.Edge;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;


/**
 * Utility functions for the class UnweightedShortestPath (shortest path 
 * distances for graphs whose edges are not weighted (using BFS)).
 * 1) Creates shortest path as a List of vertices.
 * 2) Calculates distance from the source vertex till the predefined subset of 
 * vertices, if the shortest path does not contains other vertices from 
 * the subset.
 * 
 * @author Andrew Krizhanovsky
 */
public class UnweightedShortestPathUtil {
   
    /** Gets the shortest path from source vertex to the dest as list. */
    public static List<Vertex> getPath (Vertex source,Vertex dest, UnweightedShortestPath usp) {
        
        if(null == source || null == dest) {
            return null;
        }
        
        List<Vertex> l = new ArrayList<Vertex>();
        l.add(dest);
        if(source == dest) 
            return l;
        
        Vertex v = dest;
        while (source != v) {
            Edge e = (Edge)usp.getIncomingEdgeMap(source).get(v);
            if(null == e) {
                return null;
            };
            v = e.getOpposite(v);
            l.add(v);
        }
        return l;
    }
    
    /** Calculates the length of shortest path from source till dest constrained 
     * by vertices v_constr.
     * @return -1 if (1) shortest path (Next(source),Previous(dest)) contains 
     * vertex from v_constr, or (2) there is no path
     */
    public static int getPathLen (Vertex source,Vertex dest,UnweightedShortestPath usp,
                              //Set<Vertex> v_constr)
                                List<Vertex> v_constr)
    {
        if(null == source || null == dest)
            return -1;
        int len = 0;
        if(source == dest) 
            return len;
        
        Vertex v = dest;
        while (source != v) {
            if(v != dest && v_constr.contains(v)) {
                return -1;
            }
            Edge e = (Edge)usp.getIncomingEdgeMap(source).get(v);
            if(null == e) {
                return -1;
            };
            v = e.getOpposite(v);
            len ++;
        }
        return len;
    }
    
    /** Calculates the length of shortes path from source till dest vertex.
     * @return -1 if there is no path.
     */
    public static int getPathLen (Vertex source,Vertex dest,UnweightedShortestPath usp)
    {
        if(null == source || null == dest)
            return -1;
        int len = 0;
        if(source == dest) 
            return len;
        
        Vertex v = dest;
        while (source != v) {
            //if(v != dest) {
            //    return -1;
            //}
            Edge e = (Edge)usp.getIncomingEdgeMap(source).get(v);
            if(null == e) {
                return -1;
            };
            v = e.getOpposite(v);
            len ++;
        }
        return len;
    }
    
    
    
    public static Map<Vertex,Integer> getPath (Vertex source,Set<Vertex> dest,UnweightedShortestPath usp) {
        
        List<Vertex> l = new ArrayList<Vertex>();
        l.addAll(dest);
        return UnweightedShortestPathUtil.getPath (source, l, usp);
    }
    
    public static Map<Vertex,Integer> getPathReal (Vertex source,Set<Vertex> dest,UnweightedShortestPath usp) {
        
        Map<Vertex,Integer> m = new HashMap<Vertex,Integer>();
        for(Vertex v_dest:dest) {
            int len = UnweightedShortestPathUtil.getPathLen (source, v_dest, usp);
            if(-1 != len && 0 != len) {
                m.put(v_dest, len);
            }
        }
        return m;
    }
    
    /** Gets the shortest path between sets of vertices: from the 'source' set 
     * vertices to the destination set of vertices 'dest'.
     * Attention: returned set 
     *      (1) is the ordered set of vertices.
     *      (2) does'not contain start and end vertices.
     */
    public static Vertex[] getShortestPath (Set<Vertex> source,Set<Vertex> dest,UnweightedShortestPath usp) {
        
        if(null == source || source == dest || null == dest || 
              0 == source.size()    ||            0 == dest.size())
            return null;
        
        //Set<Vertex> path_min = new HashSet<Vertex>();
        List<Vertex> path_min = new ArrayList<Vertex>();
        
        int len_min = Integer.MAX_VALUE;
        for(Vertex v_source:source) {
            for(Vertex v_dest:dest) {
                int len = -1;
                List<Vertex> path = new ArrayList<Vertex>();
                Vertex v = v_dest;
                path.add(v);
                while (v_source != v) {
                    if((v != v_source && source.contains(v)) ||
                       (v != v_dest   &&   dest.contains(v))  )
                    {
                        break;
                    }
                    Edge e = (Edge)usp.getIncomingEdgeMap(v_source).get(v);
                    if(null == e) {
                        break;
                    };
                    v = e.getOpposite(v);
                    path.add(v);
                    len ++;
                }
                
                if(-1 < len && len < len_min) {
                    len_min = len;
                    path_min.clear();
                    //path_min.add(v_source);
                    path_min.addAll(path);
                } else {
                    path.clear();
                }
                //Vertex[] ptemp = (Vertex[])path_min.toArray(new Vertex[0]);
                //int z = 0;
                //watch variables: ptemp[0].getUserDatum(GraphComparator.NAME_KEY)
            }
        }
        return (Vertex[])path_min.toArray(new Vertex[0]);
    }
    
    /** Gets the length of shortest paths from source vertex to the set 
     * of vertices dest.
     * If (1) shorthest path (source,B) contains vertex A && (2) A is vertices 
     * from dest then len for vertex B will not be stored in result Map.
     * @ return Map<vertex_j, shortest_path_len(source,vertex_j)>
     */
    public static Map<Vertex,Integer> getPath (Vertex source,List<Vertex> dest,UnweightedShortestPath usp) {
    //public static Map<Vertex,Integer> getPath (Vertex source,Set<Vertex> dest,UnweightedShortestPath usp) {
        
        Map<Vertex,Integer> m = new HashMap<Vertex,Integer>();
        
        for(Vertex v:dest) {
            int len = getPathLen (source, v, usp, dest);
            if(-1 != len && 0 != len) {
                m.put(v, len);
                
                //System.out.println("source="+(String)source.getUserDatum(GraphComparator.NAME_KEY)+
                //        "; dest=" + (String)v.getUserDatum(GraphComparator.NAME_KEY) + 
                //        "; len="+len);
            }
        }
        return m;
    }
    
    
}
